import asyncio
import pickle
import socket
import sys
from abc import ABC, abstractmethod
from logging import LogRecord
from logging.handlers import SocketHandler
from typing import Any, Callable, Dict, Literal, Optional, Union

__all__ = [
    "IPCEncoder",
    "JSONEncoder",
    "PickleEncoder",
    "SocketLogHandlerEncoder",
    "SocketServer",
    "SocketClient",
    "find_free_port",
]


class IPCEncoder(ABC):
    """An abstract base class for all encoders; implementing the necessary communication protocol requires only the definition of two abstract methods. Be mindful that varying header lengths will impact the maximum packaging size."""

    # HEAD_LENGTH: Package length
    # 1: 256 B, 2: 64 KB, 3: 16 MB, 4: 4 GB, 5: 1 TB, 6: 256 TB
    HEAD_SIZE = 4
    BYTEORDER: Literal["little", "big"] = "big"

    @abstractmethod
    def _dumps(self, raw: Any) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def _loads(self, raw: bytes) -> Any:
        raise NotImplementedError

    def dumps(self, raw: Any) -> bytes:
        result = self._dumps(raw)
        head = len(result).to_bytes(self.HEAD_SIZE, self.BYTEORDER)
        return head + result

    def loads(self, raw: Any) -> bytes:
        return self._loads(raw)

    def get_size(self, head: bytes):
        return int.from_bytes(head, self.BYTEORDER)


class JSONEncoder(IPCEncoder):
    _DUMP_KWARGS: Dict = {}
    _LOAD_KWARGS: Dict = {}
    _ENCODING = "utf-8"

    def __init__(self):
        import json

        self._encoder = json.dumps
        self._decoder = json.loads
        super().__init__()

    def _dumps(self, raw: Any) -> bytes:
        return self._encoder(raw, **self._DUMP_KWARGS).encode(self._ENCODING)

    def _loads(self, raw: bytes) -> Any:
        return self._decoder(raw.decode(self._ENCODING), **self._LOAD_KWARGS)


class PickleEncoder(IPCEncoder):
    _DUMP_KWARGS: Dict = {}
    _LOAD_KWARGS: Dict = {}

    def __init__(self):
        super().__init__()

    def _dumps(self, raw: Any) -> bytes:
        return pickle.dumps(raw, **self._DUMP_KWARGS)

    def _loads(self, raw: bytes) -> Any:
        return pickle.loads(raw, **self._LOAD_KWARGS)


class SocketLogHandlerEncoder(IPCEncoder):
    """For a practical demonstration, refer to the test code: morebuiltins/ipc.py:_test_ipc_logging.

    ```
    async def _test_ipc_logging():
        import logging

        host = "127.0.0.1"
        port = 8090
        async with SocketServer(host=host, port=port, encoder=SocketLogHandlerEncoder()):
            # socket logger demo
            # ==================
            logger = logging.getLogger("test_logger")
            logger.setLevel(logging.DEBUG)
            h = SocketHandler(host, port)
            h.setLevel(logging.DEBUG)
            logger.addHandler(h)
            logger.info("test socket")
            # ==================

            # ensure test case
            await asyncio.sleep(0.1)
            assert pickle.loads(h.sock.recv(100000)[4:])["name"] == logger.name
    ```
    And provide a simple implementation for generating logs for coroutine code with Client usage.
    """

    _DUMP_KWARGS: Dict = {"protocol": 1}
    _LOAD_KWARGS: Dict = {}

    def __init__(self):
        super().__init__()

    def makePickle(self, d: Union[dict, LogRecord]):
        if isinstance(d, LogRecord):
            msg = d.getMessage()
            d = dict(d.__dict__)
            d["msg"] = msg
            d["args"] = None
            d["exc_info"] = None
            d.pop("message", None)
        return pickle.dumps(d, **self._DUMP_KWARGS)

    def _dumps(self, raw: Union[LogRecord, dict]) -> bytes:
        return self.makePickle(raw)

    def _loads(self, raw: bytes) -> Any:
        return pickle.loads(raw, **self._LOAD_KWARGS)


class SocketServer:
    """To see an example in action, view the test code: morebuiltins/ipc.py:_test_ipc.

        ```
    async def test_client(host="127.0.0.1", port=8090, encoder=None, cases=None):
        async with SocketClient(host=host, port=port, encoder=encoder) as c:
            for case in cases:
                await c.send(case)
                response = await c.recv()
                if globals().get("print_log"):
                    print("[Client]", "send:", repr(case), "=>", "recv:", repr(response))
                assert case == response or str(case) == response, [case, response]
            await c.send("[shutdown server]")


    async def _test_ipc():
        import platform

        JSONEncoder._DUMP_KWARGS["default"] = str
        for enc, cases in [
            [PickleEncoder, [123, "123", None, {"a"}, ["a"], ("a",), {"a": 1}]],
            [JSONEncoder, [123, "123", None, {"a"}, ["a"], {"a": 1}]],
        ]:
            encoder = enc()
            if platform.system() == "Linux":
                # test unix domain socket
                print("Test Linux Unix Domain Socket")
                host = "/tmp/uds.sock"
                port = None
                async with SocketServer(host=host, port=port, encoder=encoder):
                    await test_client(host, port=None, encoder=encoder, cases=cases)

            # test socket
            host = "127.0.0.1"
            port = 8090
            async with SocketServer(host=host, port=port, encoder=encoder):
                await test_client(host="127.0.0.1", port=8090, encoder=encoder, cases=cases)


        ```
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 8090,
        handler: Optional[Callable] = None,
        encoder: Optional[IPCEncoder] = None,
        connect_kwargs: Optional[dict] = None,
    ):
        if port is None and sys.platform == "win32":
            raise SystemError("not support UDS(unix domain socket) on win32 platform")
        self.host = host
        self.port = port
        self.handler = handler or self.default_handler
        self.encoder: IPCEncoder = encoder or PickleEncoder()
        self.connect_kwargs = connect_kwargs or {}

        self.server: Optional[asyncio.base_events.Server] = None
        self._shutdown_ev: Optional[asyncio.Event] = None

    @staticmethod
    async def default_handler(self: "SocketServer", item: Any):
        if globals().get("print_log"):
            print("[Server] recv:", repr(item), "=>", "send:", repr(item), flush=True)
        if item == "[shutdown server]":
            await self.close()
        else:
            return self.encoder.dumps(item)

    async def connect_callback(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        head_size = self.encoder.HEAD_SIZE
        need_await = asyncio.iscoroutinefunction(self.handler)
        while self.is_serving() and not reader.at_eof():
            head = await reader.read(head_size)
            if len(head) < head_size:
                break
            content_length = self.encoder.get_size(head)
            # read the whole package
            head = await reader.read(content_length)
            while len(head) < content_length:
                head = head + await reader.read(content_length - len(head))
            item = self.encoder._loads(head)
            result = self.handler(self, item)
            if need_await:
                result = await result
            if isinstance(result, bytes):
                writer.write(result)
                await writer.drain()
        writer.close()
        await writer.wait_closed()

    async def close(self):
        if self.is_serving():
            self.server.close()
            await self.server.wait_closed()
            self._shutdown_ev.set()

    async def start(self):
        self._shutdown_ev = asyncio.Event()
        if self.port:
            self.server = await asyncio.start_server(
                self.connect_callback,
                host=self.host,
                port=self.port,
                **self.connect_kwargs,
            )
            await self.server.start_serving()
        else:
            self.server = await asyncio.start_unix_server(
                self.connect_callback, path=self.host, **self.connect_kwargs
            )

    def is_serving(self):
        return self.server and self.server.is_serving()

    async def wait_closed(self):
        if self._shutdown_ev:
            await self._shutdown_ev.wait()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, *_errors):
        await self.close()


class SocketClient:
    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 8090,
        encoder: Optional[IPCEncoder] = None,
        connect_kwargs: Optional[dict] = None,
    ):
        self.host = host
        self.port = port
        self.encoder: IPCEncoder = encoder or PickleEncoder()
        self.connect_kwargs = connect_kwargs or {}

    async def __aenter__(self):
        if self.port:
            self.reader, self.writer = await asyncio.open_connection(
                self.host, self.port, **self.connect_kwargs
            )
        else:
            self.reader, self.writer = await asyncio.open_unix_connection(
                path=self.host, **self.connect_kwargs
            )
        self.lock = asyncio.Lock()
        return self

    async def __aexit__(self, *_):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()

    async def send(self, item: Any):
        assert self.writer, "use `async with`"
        async with self.lock:
            self.writer.write(self.encoder.dumps(item))
            await self.writer.drain()

    async def recv(self) -> Any:
        reader = self.reader
        assert reader, "use `async with`"
        async with self.lock:
            assert not reader.at_eof()
            head = await reader.read(self.encoder.HEAD_SIZE)
            if len(head) < self.encoder.HEAD_SIZE:
                raise RuntimeError()
            content_length = self.encoder.get_size(head)
            head = await reader.read(content_length)
            while len(head) < content_length:
                head = head + await reader.read(content_length - len(head))
            return self.encoder.loads(head)


def find_free_port(host="127.0.0.1", port=0):
    """Finds and returns an available port number.

    Parameters:
    - host: The host address to bind, default is "127.0.0.1".
    - port: The port number to attempt binding, default is 0 (for OS allocation).

    Returns:
    - If a free port is found, it returns the port number; otherwise, returns None.

    Demo:

    >>> free_port = find_free_port()
    >>> isinstance(free_port, int)
    True
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((host, port))
            return s.getsockname()[1]
    except socket.error:
        pass


async def test_client(host="127.0.0.1", port=8090, encoder=None, cases=None):
    async with SocketClient(host=host, port=port, encoder=encoder) as c:
        for case in cases:
            await c.send(case)
            response = await c.recv()
            if globals().get("print_log"):
                print("[Client]", "send:", repr(case), "=>", "recv:", repr(response))
            assert case == response or str(case) == response, [case, response]
        await c.send("[shutdown server]")


async def _test_ipc():
    import platform

    JSONEncoder._DUMP_KWARGS["default"] = str
    for enc, cases in [
        [PickleEncoder, [123, "123", None, {"a"}, ["a"], ("a",), {"a": 1}]],
        [JSONEncoder, [123, "123", None, {"a"}, ["a"], {"a": 1}]],
    ]:
        encoder = enc()
        if platform.system() == "Linux":
            # test unix domain socket
            print("Test Linux Unix Domain Socket")
            host = "/tmp/uds.sock"
            port = None
            async with SocketServer(host=host, port=port, encoder=encoder):
                await test_client(host, port=None, encoder=encoder, cases=cases)

        # test socket
        host = "127.0.0.1"
        port = 8090
        async with SocketServer(host=host, port=port, encoder=encoder):
            await test_client(host="127.0.0.1", port=8090, encoder=encoder, cases=cases)


async def _test_ipc_logging():
    import logging

    host = "127.0.0.1"
    port = 8090
    assert find_free_port(host, port) == port
    async with SocketServer(host=host, port=port, encoder=SocketLogHandlerEncoder()):
        assert find_free_port(host, port) is None
        # socket logger demo
        # ==================
        logger = logging.getLogger("test_logger")
        logger.setLevel(logging.DEBUG)
        h = SocketHandler(host, port)
        h.setLevel(logging.DEBUG)
        logger.addHandler(h)
        logger.info("test socket")
        # ==================

        # ensure test case
        await asyncio.sleep(0.1)
        assert pickle.loads(h.sock.recv(100000)[4:])["name"] == logger.name
        h.sock.close()


def test():
    globals().setdefault("print_log", True)  # local test show logs
    for function in [_test_ipc, _test_ipc_logging]:
        asyncio.get_event_loop().run_until_complete(function())


if __name__ == "__main__":
    test()



import asyncio
import atexit
import base64
import gzip
import hashlib
import json
import os
import re
import subprocess
import sys
import tempfile
import traceback
from collections import UserDict
from enum import IntEnum
from functools import wraps
from itertools import groupby, islice
from os.path import basename, exists
from pathlib import Path
from time import gmtime, mktime, strftime, strptime, time, timezone
from typing import (
    Any,
    AnyStr,
    Callable,
    Dict,
    Generator,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

__all__ = [
    "ttime",
    "ptime",
    "slice_into_pieces",
    "slice_by_size",
    "unique",
    "retry",
    "guess_interval",
    "get_hash",
    "find_jsons",
    "code_inline",
    "read_size",
    "read_time",
    "Validator",
    "stagger_sort",
    "default_dict",
    "always_return_value",
    "format_error",
    "Trie",
    "GuessExt",
    "xor_encode_decode",
    "is_running",
    "set_pid_file",
    "get_paste",
    "set_clip",
]


def ttime(
    timestamp: Optional[Union[float, int]] = None,
    tzone: int = int(-timezone / 3600),
    fmt="%Y-%m-%d %H:%M:%S",
) -> str:
    """Converts a timestamp to a human-readable timestring formatted as %Y-%m-%d %H:%M:%S.

    >>> ttime(1486572818.421858323, tzone=8)
    '2017-02-09 00:53:38'

    Args:
        timestamp (float, optional): the timestamp float. Defaults to time.time().
        tzone (int, optional): time compensation. Defaults to int(-time.timezone / 3600).
        fmt (str, optional): strftime fmt. Defaults to "%Y-%m-%d %H:%M:%S".

    Returns:
        str: time string formatted.
    """
    fix_tz = tzone * 3600
    timestamp = time() if timestamp is None else timestamp
    return strftime(fmt, gmtime(timestamp + fix_tz))


def ptime(
    timestring: Optional[str] = None,
    tzone: int = int(-timezone / 3600),
    fmt: str = "%Y-%m-%d %H:%M:%S",
) -> int:
    """Converts a timestring formatted as %Y-%m-%d %H:%M:%S back into a timestamp.

    >>> ptime("2018-03-15 01:27:56", tzone=8)
    1521048476

    Args:
        timestring (str, optional): string like 2018-03-15 01:27:56. Defaults to ttime().
        tzone (int, optional): time compensation. Defaults to int(-timezone / 3600).
        fmt (_type_, optional): strptime fmt. Defaults to "%Y-%m-%d %H:%M:%S".

    Returns:
        str: time string formatted.
    """
    fix_tz = -(tzone * 3600 + timezone)
    #: str(timestring) for datetime.datetime object
    if timestring:
        return int(mktime(strptime(str(timestring), fmt)) + fix_tz)
    else:
        return int(time())


def slice_into_pieces(
    items: Sequence, n: int
) -> Generator[Union[tuple, Sequence], None, None]:
    """Divides a sequence into “n” segments, returning a generator that yields “n” pieces.

    >>> for chunk in slice_into_pieces(range(10), 3):
    ...     print(chunk)
    (0, 1, 2, 3)
    (4, 5, 6, 7)
    (8, 9)

    Args:
        seq (_type_): input a sequence.
        n (_type_): split the given sequence into "n" pieces.

    Returns:
        Generator[tuple, None, None]: a generator with tuples.

    Yields:
        Iterator[Generator[tuple, None, None]]: a tuple with n of items.
    """
    length = len(items)
    if length % n == 0:
        size = length // n
    else:
        size = length // n + 1
    for it in slice_by_size(items, size):
        yield it


def slice_by_size(
    items: Sequence, size: int, callback=tuple
) -> Generator[Union[tuple, Sequence], None, None]:
    """Slices a sequence into chunks of a specified “size”, returning a generator that produces tuples of chunks.

    >>> for chunk in slice_by_size(range(10), 3):
    ...     print(chunk)
    (0, 1, 2)
    (3, 4, 5)
    (6, 7, 8)
    (9,)

    Args:
        items (Sequence): _description_
        size (int): _description_

    Returns:
        Generator[tuple, None, None]: a generator with tuples.

    Yields:
        Iterator[Generator[tuple, None, None]]: a tuple with n of items.
    """
    iter_seq = iter(items)
    while True:
        chunk = callback(islice(iter_seq, size))
        if chunk:
            yield chunk
        else:
            break


def unique(
    items: Sequence, key: Optional[Callable] = None
) -> Generator[Any, None, None]:
    """Removes duplicate elements from a sequence while preserving the original order efficiently.

    >>> a = ['01', '1', '2']
    >>> list(unique(a, int))
    [1, 2]
    >>> list(unique(a))
    ['01', '1', '2']

    Args:
        items (Sequence): raw sequence.
        key (Callable): the function to normalize each item.

    Returns:
        Generator[tuple, None, None]: a generator with unique items.

    Yields:
        Iterator[Generator[tuple, None, None]]: the unique item.
    """
    seen: set = set()
    _add = seen.add
    if key:
        for item in items:
            _item = key(item)
            if _item not in seen:
                yield _item
                _add(_item)
    else:
        for item in items:
            if item not in seen:
                yield item
                _add(item)


def retry(
    tries=2,
    exceptions: Tuple[Type[BaseException]] = (Exception,),
    return_exception=False,
):
    """A decorator that retries the decorated function up to “tries” times if the specified exceptions are raised.

    >>> func = lambda items: 1/items.pop(0)
    >>> items = [0, 1]
    >>> new_func = retry(tries=2, exceptions=(ZeroDivisionError,))(func)
    >>> new_func(items)
    1.0

    Args:
        tries (int, optional): try n times, if n==1 means no retry. Defaults to 1.
        exceptions (Tuple[Type[BaseException]], optional): only retry the given errors. Defaults to (Exception,).
        return_exception (bool, optional): raise the last exception or return it. Defaults to False.
    """

    def wrapper(function):
        @wraps(function)
        def retry_sync(*args, **kwargs):
            for _ in range(tries):
                try:
                    return function(*args, **kwargs)
                except exceptions as err:
                    error = err
            if return_exception:
                return error
            raise error

        @wraps(function)
        async def retry_async(*args, **kwargs):
            for _ in range(tries):
                try:
                    return await function(*args, **kwargs)
                except exceptions as err:
                    error = err
            if return_exception:
                return error
            raise error

        if asyncio.iscoroutinefunction(function):
            return retry_async
        else:
            return retry_sync

    return wrapper


def guess_interval(nums, accuracy=0):
    """Analyzes a sequence of numbers and returns the median, calculating intervals only if they are greater than or equal to the specified accuracy.

    >>> # sorted_seq: [2, 10, 12, 19, 19, 29, 30, 32, 38, 40, 41, 54, 62]
    >>> # diffs: [8, 7, 10, 6, 13, 8]
    >>> # median: 8
    >>> seq = [2, 10, 12, 19, 19, 29, 30, 32, 38, 40, 41, 54, 62]
    >>> guess_interval(seq, 5)
    8

    """
    if not nums:
        return 0
    nums = sorted([int(i) for i in nums])
    if len(nums) == 1:
        return nums[0]
    diffs = [nums[i + 1] - nums[i] for i in range(len(nums) - 1)]
    diffs = [item for item in diffs if item >= accuracy]
    sorted_diff = sorted(diffs)
    return sorted_diff[len(diffs) // 2]


def get_hash(
    string,
    n: Union[tuple, list, int, None] = None,
    default: Callable = lambda obj: str(obj).encode("utf-8"),
    func=hashlib.md5,
) -> str:
    """Generates an MD5 hash string from the given input string.

    >>> get_hash(123456, 10)
    'a59abbe56e'
    >>> get_hash('test')
    '098f6bcd4621d373cade4e832627b4f6'
    >>> get_hash(['list_demo'], (5, 10))
    '7152a'
    >>> get_hash(['list_demo'], func=hashlib.sha256)
    'a6072e063d36a09052a9e5eb389a425a3dc158d3a5955808159a118aa192c718'
    >>> get_hash(['list_demo'], 16, func=hashlib.sha256)
    '389a425a3dc158d3'
    """
    if isinstance(string, bytes):
        _bytes = string
    else:
        _bytes = default(string)
    _temp = func(_bytes).hexdigest()
    if n is None:
        return _temp
    elif isinstance(n, (int, float)):
        start, end = (len(_temp) - n) // 2, (n - len(_temp)) // 2
        return _temp[start:end]
    elif isinstance(n, (tuple, list)):
        start, end = n[0], n[1]
        return _temp[start:end]


def find_jsons(
    string,
    return_as: Literal["json", "object", "index"] = "json",
    json_loader=json.loads,
):
    """A generator that locates valid JSON strings, supporting only dictionaries and lists.

    >>> list(find_jsons('string["123"]123{"a": 1}[{"a": 1, "b": [1,2,3]}]'))
    ['["123"]', '{"a": 1}', '[{"a": 1, "b": [1,2,3]}]']
    >>> list(find_jsons('string[]{}{"a": 1}'))
    ['[]', '{}', '{"a": 1}']
    >>> list(find_jsons('string[]|{}string{"a": 1}', return_as='index'))
    [(6, 8), (9, 11), (17, 25)]
    >>> list(find_jsons('xxxx[{"a": 1, "b": [1,2,3]}]xxxx', return_as='object'))
    [[{'a': 1, 'b': [1, 2, 3]}]]
    """

    def find_matched(string, left, right):
        _stack = []
        for index, char in enumerate(string):
            if char == left:
                _stack.append(index)
            elif char == right:
                try:
                    _stack.pop()
                except IndexError:
                    break
            else:
                continue
            if not _stack:
                return index

    search = re.search
    brackets_map = {"{": "}", "[": "]"}
    current_start = 0
    while string and isinstance(string, str):
        _match = search(r"[\[\{]", string)
        if not _match:
            break
        left = _match.group()
        right = brackets_map[left]
        _start = _match.span()[0]
        sub_string = string[_start:]
        _end = find_matched(sub_string, left, right)
        if _end is None:
            # not found matched, check next left
            string = sub_string
            continue
        string = sub_string[_end + 1 :]
        try:
            _partial = sub_string[: _end + 1]
            _loaded_result = json_loader(_partial)
            yield {
                "json": _partial,
                "object": _loaded_result,
                "index": (current_start + _start, current_start + _start + _end + 1),
            }.get(return_as, string)
        except (ValueError, TypeError):
            pass
        current_start += _start + _end + 1


def code_inline(
    source_code: str,
    encoder: Literal["b16", "b32", "b64", "b85"] = "b85",
) -> str:
    """Minifies Python source code into a single line.

    >>> code1 = ''
    >>> code2 = code_inline("variable=12345")
    >>> # import base64,gzip;exec(gzip.decompress(base64.b85decode("ABzY8mBl+`0{<&ZEXqtw%1N~~G%_|Z1ptx!(o_xr000".encode("u8"))))
    >>> exec(code2)
    >>> variable
    12345

    Args:
        source_code (str): python original code.
        encoder (Literal['b16', 'b32', 'b64', 'b85'], optional): base64.encoder. Defaults to "b85".
    Returns:
        new source code inline.
    """
    _encoder = getattr(base64, f"{encoder}encode")
    _source = source_code.encode(encoding="u8")
    _source = gzip.compress(_source)
    _source = _encoder(_source)
    code = _source.decode("u8")
    result = (
        "import base64,gzip;exec(gzip.decompress("
        f'base64.{encoder}decode("{code}".encode("u8"))))'
    )
    return result


class BytesUnit(IntEnum):
    B = 0
    KB = 1 * 1024**1
    MB = 1 * 1024**2
    GB = 1 * 1024**3
    TB = 1 * 1024**4
    PB = 1 * 1024**5
    EB = 1 * 1024**6
    ZB = 1 * 1024**7
    YB = 1 * 1024**8


class TimeUnit(IntEnum):
    secs = 0
    mins = 60
    hours = 60 * 60
    days = 60 * 60 * 24
    mons = 60 * 60 * 24 * 30
    years = 60 * 60 * 24 * 365


def read_num(
    b,
    enum=IntEnum,
    rounded: Optional[int] = None,
    shorten=False,
    precision=1.0,
    sep=" ",
):
    units = tuple(enum)
    unit = units[0]
    ntime = 0.5 if rounded and rounded > 0 else 1
    for _unit in units:
        if b >= _unit.value * ntime:
            unit = _unit
        else:
            break
    result = round(b / (unit.value or 1), rounded)
    if shorten and isinstance(result, float):
        int_result = int(result)
        if int_result / result >= precision:
            result = int_result
    return f"{result}{sep}{unit.name}"


def read_size(b, rounded: Optional[int] = None, shorten=False, precision=1.0, sep=" "):
    """Converts byte counts into a human-readable string. Setting shorten=True and precision=0.99 will trim unnecessary decimal places from the tail of floating-point numbers.

    >>> (read_size(1023), read_size(1024))
    ('1023 B', '1 KB')
    >>> (read_size(400.5, 1), read_size(400.5, 1, True), read_size(400.5, 1, True, 0.99))
    ('400.5 B', '400.5 B', '400 B')
    >>> (read_size(511.55, 1, shorten=True, precision=0.999), read_size(512.55, 1, shorten=True, precision=0.999))
    ('511.6 B', '0.5 KB')
    >>> (read_size(511, 1, shorten=True), read_size(512, 1, shorten=True))
    ('511 B', '0.5 KB')
    >>> read_size(512, 1, sep='')
    '0.5KB'
    >>> read_size(1025, 1, shorten=False), read_size(1025, 1, shorten=True)
    ('1.0 KB', '1 KB')
    >>> for i in range(0, 5):
    ...     [1.1111 * 1024**i, i, read_size(1.1111 * 1024**i, rounded=i)]
    ...
    [1.1111, 0, '1.0 B']
    [1137.7664, 1, '1.1 KB']
    [1165072.7936, 2, '1.11 MB']
    [1193034540.6464, 3, '1.111 GB']
    [1221667369621.9136, 4, '1.1111 TB']

    Args:
        b: bytes
        rounded (int, optional): arg for round. Defaults to None.
        shorten (bool): shorten unnecessary tail 0.
        precision: (float): shorten precision, often set to 0.99.
        sep (str, optional): sep between result and unit.

    Returns:
        str


    """
    return read_num(
        b,
        BytesUnit,
        rounded=rounded,
        shorten=shorten,
        precision=precision,
        sep=sep,
    )


def read_time(
    secs, rounded: Optional[int] = None, shorten=False, precision=1.0, sep=" "
):
    """Converts seconds into a more readable time duration string.

    >>> read_time(0)
    '0 secs'
    >>> read_time(60)
    '1 mins'
    >>> for i in range(0, 6):
    ...     [1.2345 * 60**i, read_time(1.2345 * 60**i, rounded=1)]
    ...
    [1.2345, '1.2 secs']
    [74.07, '1.2 mins']
    [4444.2, '1.2 hours']
    [266652.0, '3.1 days']
    [15999120.0, '0.5 years']
    [959947200.0, '30.4 years']

    Args:
        b: seconds
        rounded (int, optional): arg for round. Defaults to None.
        shorten (bool): shorten unnecessary tail 0.
        precision: (float): shorten precision, often set to 0.99.
        sep (str, optional): sep between result and unit.

    Returns:
        str
    """
    return read_num(
        secs,
        TimeUnit,
        rounded=rounded,
        shorten=shorten,
        precision=precision,
        sep=sep,
    )


class Validator:
    """Validator for dataclasses.
    >>> from dataclasses import dataclass, field
    >>>
    >>>
    >>> @dataclass
    ... class Person(Validator):
    ...     screen: dict = field(metadata={"callback": lambda i: i["s"]})
    ...     name: str = field(default=None, metadata={"callback": str})
    ...     age: int = field(default=0, metadata={"callback": int})
    ...
    >>>
    >>> print(Person({"s": 3}, 123, "123"))
    Person(screen=3, name='123', age=123)
    """

    def __post_init__(self):
        for f in self.__dataclass_fields__.values():
            callback = f.metadata.get("callback")
            if callback:
                setattr(self, f.name, callback(getattr(self, f.name)))

    # def to_dict(self):
    #     return {
    #         field.name: getattr(self, field.name)
    #         for field in self.__dataclass_fields__.values()
    #     }


def stagger_sort(items, group_key, sort_key=None):
    """Ensures that identical groups are ordered and evenly distributed, mitigating data skew. The function does not alter the original list and returns a generator.

    >>> items = [('a', 0), ('a', 2), ('a', 1), ('b', 0), ('b', 1)]
    >>> list(stagger_sort(items, sort_key=lambda i: (i[0], i[1]), group_key=lambda i: i[0]))
    [('a', 0), ('b', 0), ('a', 1), ('b', 1), ('a', 2)]

    """
    if sort_key:
        items = sorted(items, key=sort_key)
    buckets = [list(group[1]) for group in groupby(items, group_key)]
    while True:
        next_buckets = []
        for items in buckets:
            try:
                yield items.pop(0)
                next_buckets.append(items)
            except IndexError:
                pass
        if next_buckets:
            buckets = next_buckets
        else:
            break


def default_dict(cls: Type[dict], **kwargs) -> dict:
    """Initializes a dictionary with default zero values based on a subclass of TypedDict.

    >>> class Demo(dict):
    ...     int_obj: int
    ...     float_obj: float
    ...     bytes_obj: bytes
    ...     str_obj: str
    ...     list_obj: list
    ...     tuple_obj: tuple
    ...     set_obj: set
    ...     dict_obj: dict
    >>> default_dict(Demo, bytes_obj=b'1')
    {'int_obj': 0, 'float_obj': 0.0, 'bytes_obj': b'1', 'str_obj': '', 'list_obj': [], 'tuple_obj': (), 'set_obj': set(), 'dict_obj': {}}
    """
    result = cls()
    built_in_types = {int, float, bytes, str, list, tuple, set, dict}
    for key, tp in cls.__annotations__.items():
        if key in kwargs:
            result[key] = kwargs[key]
        elif tp in built_in_types:
            result[key] = tp()
    return result


def always_return_value(value):
    """Got a function always return the given value.
    >>> func = always_return_value(1)
    >>> func(1, 2, 3)
    1
    >>> func(1, 2, c=3)
    1
    """

    def func(*args, **kws):
        return value

    return func


def _tb_filter(tb: traceback.FrameSummary):
    # filt tb with code-line and not in site-packages
    return tb.line and "site-packages" not in tb.filename


def format_error(
    error: BaseException,
    index: Union[int, slice] = slice(-3, None, None),
    filter: Optional[Callable] = _tb_filter,
    template="[{trace_routes}] {error_line} >>> {error.__class__.__name__}({error!s:.100})",
    **kwargs,
) -> str:
    r"""Extracts frame information from an exception, with an option to filter out “site-packages” details by default.

    Parameters:

    - `error` (`BaseException`): The exception instance for which the stack trace information is to be extracted and formatted.
    - `index` (`Union[int, slice]`, optional): Specifies which frames to include in the output. By default, it's set to `slice(-3, None, None)`, showing the last three frames. Can be an integer for a single frame or a slice object for a range of frames.
    - `filter` (`Optional[Callable]`, optional): A callable that determines whether a given frame should be included. Defaults to `_tb_filter`, which typically filters out frames from "site-packages". If set to `None`, no filtering occurs.
    - `template` (`str`, optional): A string template defining how the error message should be formatted. It can include placeholders like `{trace_routes}`, `{error_line}`, and `{error.__class__.__name__}`. The default template provides a concise summary of the error location and type.
    - `**kwargs`: Additional keyword arguments to be used within the formatting template.

    Returns:

    A string representing the formatted error message based on the provided parameters and template.

    Demo:

    >>> try:
    ...     # test default
    ...     1 / 0
    ... except Exception as e:
    ...     format_error(e)
    '[<doctest>:<module>:3] 1 / 0 >>> ZeroDivisionError(division by zero)'
    >>> try:
    ...     # test in function
    ...     def func1(): 1 / 0
    ...     func1()
    ... except Exception as e:
    ...     format_error(e)
    '[<doctest>:<module>:4 | <doctest>:func1:3] def func1(): 1 / 0 >>> ZeroDivisionError(division by zero)'
    >>> try:
    ...     # test index
    ...     def func2(): 1 / 0
    ...     func2()
    ... except Exception as e:
    ...     format_error(e, index=0)
    '[<doctest>:<module>:4] func2() >>> ZeroDivisionError(division by zero)'
    >>> try:
    ...     # test slice index
    ...     def func2(): 1 / 0
    ...     func2()
    ... except Exception as e:
    ...     format_error(e, index=slice(-1, None, None))
    '[<doctest>:func2:3] def func2(): 1 / 0 >>> ZeroDivisionError(division by zero)'
    >>> try:
    ...     # test with default filter(filename skip site-packages)
    ...     from pip._internal.utils.compatibility_tags import version_info_to_nodot
    ...     version_info_to_nodot(0)
    ... except Exception as e:
    ...     format_error(e)
    "[<doctest>:<module>:4] version_info_to_nodot(0) >>> TypeError('int' object is not subscriptable)"
    >>> try:
    ...     # test without filter
    ...     from pip._internal.utils.compatibility_tags import version_info_to_nodot
    ...     version_info_to_nodot(0)
    ... except Exception as e:
    ...     format_error(e, filter=None)
    '[<doctest>:<module>:4 | compatibility_tags.py:version_info_to_nodot:23] return "".join(map(str, version_info[:2])) >>> TypeError(\'int\' object is not subscriptable)'
    >>> try:
    ...     # test with custom filter.
    ...     from pip._internal.utils.compatibility_tags import version_info_to_nodot
    ...     version_info_to_nodot(0)
    ... except Exception as e:
    ...     format_error(e, filter=lambda i: '<doctest' in str(i))
    "[<doctest>:<module>:4] version_info_to_nodot(0) >>> TypeError('int' object is not subscriptable)"
    """
    try:
        filter = filter or always_return_value(True)
        tbs = [tb for tb in traceback.extract_tb(error.__traceback__) if filter(tb)]
        if isinstance(index, slice):
            tbs = tbs[index]
        elif isinstance(index, int):
            tbs = [tbs[index]]
        else:
            raise ValueError("Invalid index type")
        trace_route_list = []
        for tb in tbs:
            filename = tb.filename
            if exists(filename):
                _basename = basename(filename)
            elif filename[0] == "<":
                _basename = f"{filename.split()[0]}>"
            else:
                _basename = filename
            trace_route_list.append(f"{_basename}:{tb.name}:{tb.lineno}")
        trace_routes = " | ".join(trace_route_list)
        _kwargs = {
            "tbs": tbs,
            "error": error,
            "error_line": tbs[-1].line,
            "trace_routes": trace_routes,
        }
        _kwargs.update(kwargs)
        return template.format_map(_kwargs)
    except IndexError:
        return ""


class Trie(UserDict):
    """Transforms a standard dictionary into a trie structure that supports prefix matching.

    >>> trie = Trie({"ab": 1, "abc": 2, b"aa": 3, ("e", "e"): 4, (1, 2): 5})
    >>> trie
    {'a': {'b': {'_VALUE': 1, 'c': {'_VALUE': 2}}}, 97: {97: {'_VALUE': 3}}, 'e': {'e': {'_VALUE': 4}}, 1: {2: {'_VALUE': 5}}}
    >>> trie.match("abcde")
    (3, 2)
    >>> trie.match("abde")
    (2, 1)
    >>> trie.match(b"aabb")
    (2, 3)
    >>> trie.match("eee")
    (2, 4)
    >>> trie.match(list("eee"))
    (2, 4)
    >>> trie.match(tuple("eee"))
    (2, 4)
    >>> trie.match((1, 2, 3, 4))
    (2, 5)
    """

    def __init__(self, data: Dict[AnyStr, Any], value_node="_VALUE"):
        # data should be 1-depth dict
        super().__init__(self.parse_dict(data, value_node))
        self.value_node = value_node
        self.not_set = object()

    @staticmethod
    def parse_dict(data: dict, value_node="_VALUE") -> dict:
        "Convert a regular dictionary to a prefix tree dictionary"
        result: dict = {}
        for word, value in data.items():
            trie = result
            for char in word:
                if char not in trie:
                    trie[char] = {}
                trie = trie[char]
            trie[value_node] = value
        return result

    def match(self, seq, default: Tuple[Union[None, int], Any] = (None, None)):
        """Use the input sequence to match the hit value in the prefix tree as much as possible.

        Args:
            seq: iterable object
            default:return while not match. Defaults to (None, None).

        Returns:
            _type_: tuple
        """
        trie = self
        value_node = self.value_node
        value = self.not_set
        dep = 0
        for char in seq:
            if char in trie:
                dep += 1
                trie = trie[char]
                try:
                    value = trie[value_node]
                except KeyError:
                    pass
            else:
                break
        if value is not self.not_set:
            return (dep, value)
        return default


class GuessExt(object):
    r"""Determines whether the input bytes of a file prefix indicate a compressed file format.

    >>> cg = GuessExt()
    >>> cg.get_ext(b"PK\x05\x06zipfiledemo")
    '.zip'
    """

    MAGIC = {
        b"PK\x03\x04": ".zip",
        b"PK\x03\x06": ".zip",
        b"PK\x03\x08": ".zip",
        b"PK\x05\x04": ".zip",
        b"PK\x05\x06": ".zip",
        b"PK\x05\x08": ".zip",
        b"PK\x07\x04": ".zip",
        b"PK\x07\x06": ".zip",
        b"PK\x07\x08": ".zip",
        b'"\xb5/\xfd': ".zst",
        b"#\xb5/\xfd": ".zst",
        b"$\xb5/\xfd": ".zst",
        b"%\xb5/\xfd": ".zst",
        b"&\xb5/\xfd": ".zst",
        b"'\xb5/\xfd": ".zst",
        b"(\xb5/\xfd": ".zst",
        b"LZIP": ".lz",
        b"MSCF": ".cab",
        b"ISc(": ".cab",
        b"\x1f\x8b\x08": ".gz",
        b"BZh": ".bz2",
        b"\xfd7zXZ\x00": ".xz",
        b"7z\xbc\xaf'\x1c": ".7z",
        b"Rar!\x1a\x07\x00": ".rar",
        b"Rar!\x1a\x07\x01": ".rar",
    }

    def __init__(self):
        self.trie = Trie(self.MAGIC)

    def get_ext(self, magic_bytes: bytes, default=""):
        dep, value = self.trie.match(magic_bytes)
        if dep is None:
            return default
        else:
            return value


def xor_encode_decode(data, key):
    r"""
    Perform XOR encryption or decryption on the given data using a provided key.

    This function encrypts or decrypts the data by performing an XOR operation
    between each byte of the data and the corresponding byte of the key. The key
    is repeated if necessary to cover the entire length of the data.

    Parameters:
    - data: The byte data to be encrypted or decrypted.
    - key: The key used for encryption or decryption, also a sequence of bytes.

    Returns:
    - The resulting byte data after encryption or decryption.

    Example:

        >>> original_data = b'Hello, World!'
        >>> key = b'secret'
        >>> encrypted_data = xor_encode_decode(original_data, key)
        >>> encrypted_data
        b';\x00\x0f\x1e\nXS2\x0c\x00\t\x10R'
        >>> decrypted_data = xor_encode_decode(encrypted_data, key)
        >>> decrypted_data == original_data
        True
    """
    # Extend the key to ensure its length is at least as long as the data
    extended_key = key * (len(data) // len(key) + 1)
    # Perform XOR operation between each byte of the data and the extended key,
    # and return the new sequence of bytes
    return bytes([b ^ k for b, k in zip(data, extended_key)])


def is_running(pid):
    """Check if the given process ID is running.

    Parameters:
    pid -- The process ID to check.

    Returns:
    True if the process is running; False otherwise.

    Examples:
    >>> is_running(os.getpid() * 10)  # Assume process is not running
    False
    >>> is_running(os.getpid())  # Check if the current process is running
    True
    >>> is_running("not_a_pid")  # Invalid PID input should be handled and return False
    False

    """
    try:
        pid = int(pid)
    except ValueError:
        return False
    if sys.platform == "win32":
        with os.popen('tasklist /fo csv /fi "pid eq %s"' % int(pid)) as f:
            f.readline()
            text = f.readline()
            return bool(text)
    else:
        try:
            os.kill(int(pid), 0)
            return True
        except OSError:
            return False
        except SystemError:
            # maybe windows?
            return True


def set_pid_file(
    path: Optional[Union[str, Path]] = None,
    raise_error=False,
    default_dir="",
    default_name="",
    default_level=2,
):
    """Sets a PID file to prevent multiple instances of a script or process from running concurrently.
    If no path is specified, it constructs a default based on the calling script's location and naming convention.

    The function checks for an existing PID file and verifies if the process is still running.

    If the process is running and `raise_error` is True, it raises a RuntimeError; otherwise, it returns False.
    Upon successful setup, it writes the current process ID (PID) to the file and schedules the file for deletion upon exit.

    Parameters:
    - path (Optional[Union[str, Path]]): The desired path for the PID file. Defaults to a generated path based on caller.
    - raise_error (bool): Determines whether to raise an error if the PID file is already locked. Defaults to False.
    - default_dir (str): The default directory for the PID file if none is provided. Defaults to "/dev/shm" or temp dir.
    - default_name (str): The default base name for the PID file. Automatically generated with the caller filename if not provided.
    - default_level (int): The number of directory levels to include in the default name from the caller's path.

    Returns:
    - Path: The path of the PID file if successfully created or updated.
    - False: If the PID file exists and the associated process is running, and `raise_error` is False.

    Raises:
    - ValueError: If unable to determine the caller's path or set a default name for the PID file.
    - RuntimeError: If `raise_error` is True and the PID file is locked by another process.

    Examples:

    >>> path = set_pid_file()  # Assuming this is the first run, should succeed
    >>> path.name.endswith('doctest.py.pid')
    True
    >>> set_pid_file()  # Simulating second instance trying to start, should raise error if raise_error=True
    False
    >>> path.unlink()
    """
    if path is None:
        if not default_name:
            try:
                stack = traceback.extract_stack()
                for index in range(len(stack) - 2, -1, -1):
                    caller_path = Path(stack[index].filename)
                    # print(caller_path)
                    if caller_path.is_file():
                        break
                else:
                    raise ValueError("Could not find caller path")
                if not caller_path.is_file():
                    pass
                parts: list = []
                for index, part in enumerate(caller_path.as_posix().split("/")[::-1]):
                    if index < default_level:
                        parts.insert(0, part)
                    else:
                        break
                default_name = "__".join(parts) + ".pid"
            except Exception as e:
                raise ValueError("Could not set default name for PID file: %s" % e)
        if not default_dir:
            default_dir = Path("/dev/shm")
            if not default_dir.is_dir():
                default_dir = Path(tempfile.gettempdir())
        path = Path(default_dir) / default_name
    if not isinstance(path, Path):
        path = Path(path)
    running = False
    if path.is_file():
        try:
            old_pid = int(path.read_text().strip())
            running = is_running(old_pid)
        except ValueError:
            # NaN pid
            pass

    if running:
        if raise_error:
            raise RuntimeError(f"{path.as_posix()} is locked by {old_pid}")
        else:
            return False
    else:
        pid_str = str(os.getpid())
        path.write_text(pid_str)

        def _release_file():
            if path.is_file() and path.read_text() == pid_str:
                path.unlink()

        atexit.register(_release_file)
        return path


def get_paste() -> Union[str, None]:
    """This module offers a simple utility for retrieving text from the system clipboard with tkinter.

    Function:
        get_paste() -> Union[str, None]

    Usage Note:
        While this function handles basic clipboard retrieval, for more advanced scenarios such as setting clipboard content or maintaining a persistent application interface, consider using libraries like `pyperclip` or running `Tkinter.mainloop` which keeps the GUI event loop active.
        Set clipboard with tkinter:
            _tk.clipboard_clear()
            _tk.clipboard_append(text)
            _tk.update()
            _tk.mainloop() # this is needed
    """
    from tkinter import TclError, Tk

    _tk = Tk()
    _tk.withdraw()
    text = None
    try:
        text = _tk.clipboard_get()
    except TclError:
        pass
    finally:
        _tk.destroy()
        return text


def set_clip(text: str):
    """Copies the given text to the clipboard using a temporary file in a Windows environment.

    This function writes the provided text into a temporary file and then uses the `clip.exe` command-line utility
    to read from this file and copy its content into the clipboard.

    Parameters:
    text: str - The text content to be copied to the clipboard.
    """
    if sys.platform != "win32":
        raise RuntimeError("set_clip is only supported on Windows")
    try:
        path = Path(tempfile.gettempdir()) / "set_clip.txt"
        path.write_text(text)
        subprocess.run(
            f'clip.exe < "{path.name}"',
            cwd=path.parent.absolute(),
            shell=True,
        )
    finally:
        path.unlink(missing_ok=True)


if __name__ == "__main__":
    __name__ = "morebuiltins.utils"
    import doctest

    doctest.testmod()



import http.client
import json
import time
import traceback
import urllib.request
from collections import deque
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Asset(object):
    name: str
    keywords: str
    url: str
    size: int


def get_assets():
    api = (
        "https://api.github.com/repos/indygreg/python-build-standalone/releases/latest"
    )
    req = urllib.request.Request(api, headers={"User-Agent": "Chrome"})
    with urllib.request.urlopen(url=req, timeout=10) as resp:
        data = json.loads(resp.read())
        # print(data)
        urls = [
            Asset(
                i["name"],
                [k.replace(f"+{data['name']}", "") for k in i["name"].split("-", 6)],
                i["browser_download_url"],
                i["size"],
            )
            for i in data["assets"]
            if not i["name"].endswith(".sha256") and i["name"].count("-") >= 6
        ]
        return urls


def get_time():
    return time.strftime("%H:%M:%S")


def download_python():
    """Download python portable interpreter from https://github.com/indygreg/python-build-standalone/releases. `python -m morebuiltins.download_python`

    λ python -m morebuiltins.download_python
    [10:56:17] Checking https://api.github.com/repos/indygreg/python-build-standalone/releases/latest
    [10:56:19] View the rules:
    https://gregoryszorc.com/docs/python-build-standalone/main/running.html#obtaining-distributions

    [10:56:19] Got 290 urls from github.

    [290] Enter keywords (can be int index or partial match, defaults to 0):
    0. windows
    1. linux
    2. darwin
    0
    [10:56:24] Filt with keyword: "windows". 290 => 40

    [40] Enter keywords (can be int index or partial match, defaults to 0):
    0. 3.12.3
    1. 3.11.9
    2. 3.10.14
    3. 3.9.19
    4. 3.8.19

    [10:56:25] Filt with keyword: "3.12.3". 40 => 8

    [8] Enter keywords (can be int index or partial match, defaults to 0):
    0. x86_64
    1. i686

    [10:56:28] Filt with keyword: "x86_64". 8 => 4

    [4] Enter keywords (can be int index or partial match, defaults to 0):
    0. shared-pgo-full.tar.zst
    1. shared-install_only.tar.gz
    2. pgo-full.tar.zst
    3. install_only.tar.gz
    3
    [10:56:33] Filt with keyword: "install_only.tar.gz". 4 => 1
    [10:56:33] Download URL: 39.1 MB
    https://github.com/indygreg/python-build-standalone/releases/download/20240415/cpython-3.12.3%2B20240415-x86_64-pc-windows-msvc-install_only.tar.gz
    File path to save(defaults to `./cpython-3.12.3+20240415-x86_64-pc-windows-msvc-install_only.tar.gz`)?
    or `q` to exit.

    [10:56:38] Start downloading...
    https://github.com/indygreg/python-build-standalone/releases/download/20240415/cpython-3.12.3%2B20240415-x86_64-pc-windows-msvc-install_only.tar.gz
    D:\github\morebuiltins\morebuiltins\download_python\cpython-3.12.3+20240415-x86_64-pc-windows-msvc-install_only.tar.gz
    [10:56:44] Downloading: 39.12 / 39.12 MB | 100.00% | 11.3 MB/s | 0s
    [10:56:44] Download complete."""
    print(
        f"[{get_time()}] Checking https://api.github.com/repos/indygreg/python-build-standalone/releases/latest",
        flush=True,
    )
    assets = get_assets()
    print(
        f"[{get_time()}] View the rules:\nhttps://gregoryszorc.com/docs/python-build-standalone/main/running.html#obtaining-distributions\n",
        flush=True,
    )
    print(f"[{get_time()}] Got {len(assets)} urls from github.")

    def sort_key(s):
        try:
            return tuple(map(int, s.split(".")))
        except ValueError:
            return s

    indexs = [4, 1, 5, 2, 3, 0, 6]
    for index in indexs:
        try:
            to_filt = sorted(
                {i.keywords[index] for i in assets}, key=sort_key, reverse=True
            )
        except IndexError:
            continue
        if len(to_filt) == 1:
            continue
        choices = "\n".join((f"{idx}. {ii}" for idx, ii in enumerate(to_filt, 0)))
        arg = input(
            f"\n[{len(assets)}] Enter keywords (can be int index or partial match, defaults to 0):\n{choices}\n"
        )
        if not arg:
            arg = to_filt[0]
        elif arg.isdigit():
            arg = to_filt[int(arg)]
        old = len(assets)
        temp = [i for i in assets if arg == i.keywords[index]]
        if temp:
            assets = temp
        else:
            assets = [i for i in assets if arg in i.keywords[index]]
        print(
            f'[{get_time()}] Filt with keyword: "{arg}".',
            old,
            "=>",
            len(assets),
            flush=True,
        )
    while len(assets) > 1:
        names = "\n".join(i.name for i in assets)
        arg = input(f"Enter keyword to reduce the list (partial match):\n{names}\n")
        assets = [i for i in assets if arg in i.name]
    if not assets:
        input("No match, press enter to exit.")
        return
    asset = assets[0]
    download_url = asset.url
    total_size = asset.size
    print(
        f"[{get_time()}] Download URL:",
        round(total_size / 1024**2, 1),
        "MB",
    )
    print(download_url, flush=True)
    target = (
        input(
            f"File path to save(defaults to `./{asset.name}`)?\nor `q` to exit.\n"
        ).lower()
        or asset.name
    )
    if target == "q":
        return
    target_path = Path(target)
    try:
        target_path.unlink()
    except FileNotFoundError:
        pass
    print(f"[{get_time()}] Start downloading...")
    print(download_url)
    print(target_path.absolute(), flush=True)
    records = deque(maxlen=1000)

    def reporthook(blocknum, blocksize, totalsize):
        if totalsize < 0:
            totalsize = total_size
        _done = blocknum * blocksize
        if not _done:
            return
        percent = 100.0 * _done / totalsize
        total = totalsize / 1024 / 1024
        done = _done / 1024 / 1024 or total
        if percent > 100:
            percent = 100
        now = time.time()
        record = (now, _done)
        records.appendleft(record)
        timeleft = "-"
        _speed = 0
        if len(records) >= 2:
            for record in records:
                if now - record[0] > 1:
                    break
            time_diff = now - record[0]
            if time_diff:
                _speed = round((_done - record[1]) / time_diff, 1)
                secs = (totalsize - _done) / _speed
                if secs > 60:
                    timeleft = f"{int(secs / 60)}:{int(secs % 60):02}"
                else:
                    timeleft = f"{int(secs)}s"
        if _speed > 1024**2:
            speed = f"{round(_speed / 1024**2, 1)} MB/s"
        elif _speed > 1024:
            speed = f"{round(_speed / 1024, 1)} KB/s"
        else:
            speed = f"{round(_speed, 1)} B/s"
        print(
            f"[{get_time()}] Downloading: {done:.2f} / {total:.2f} MB | {percent:.2f}% | {speed} | {timeleft} {' ' * 10}",
            end="\r",
            flush=True,
        )

    temp_path = target_path.with_suffix(".tmp")
    for _ in range(3):
        try:
            urllib.request.urlretrieve(
                download_url, temp_path.absolute().as_posix(), reporthook=reporthook
            )
            temp_path.rename(target_path)
            print(
                f"\n[{get_time()}] Download complete.",
                flush=True,
            )
            break
        except http.client.RemoteDisconnected:
            continue
        except KeyboardInterrupt:
            try:
                temp_path.unlink()
            except FileNotFoundError:
                pass
            print()
            print(f"\n[{get_time()}] Download canceled.", flush=True)
            return
        except Exception:
            print()
            traceback.print_exc()
            try:
                temp_path.unlink()
            except FileNotFoundError:
                pass
            break
    print("Press enter to exit.", flush=True)
    input()


if __name__ == "__main__":
    download_python()



# -*- coding: utf-8 -*-

import compileall
import json
import re
import shutil
import sys
import tempfile
import time
import typing
import zipapp
from glob import glob
from hashlib import md5
from pathlib import Path
from pkgutil import get_data
from zipfile import ZIP_DEFLATED, ZIP_STORED, BadZipFile, ZipFile

__version__ = "2024.06.04"


def get_pip_main(ensurepip_root=None):
    try:
        import pip
    except ImportError:
        import ensurepip

        assert ensurepip._bootstrap(root=ensurepip_root) == 0
        if ensurepip_root:
            for _path in Path(ensurepip_root).glob("**/pip/"):
                if _path.is_dir():
                    sys.path.append(str(_path.parent.absolute()))
                    break
        import pip
    try:
        from pip._internal.cli.main import main

        return main
    except ImportError:
        pass
    try:
        from pip import main

        return main
    except ImportError:
        pass
    try:
        from pip._internal import main

        return main
    except ImportError:
        pass
    return pip.main


class ZipApp(object):
    DEFAULT_OUTPUT_PATH = "app.pyz"
    DEFAULT_UNZIP_CACHE_PATH = "zipapps_cache"
    AUTO_FIX_UNZIP_KEYS = {"AUTO_UNZIP", "AUTO"}
    COMPILE_KWARGS: typing.Dict[str, typing.Any] = {}
    HANDLE_OTHER_ENVS_FLAG = "--zipapps"
    LAZY_PIP_DIR_NAME = "_zipapps_lazy_pip"
    PATH_SPLIT_TAG = ","
    HANDLE_ACTIVATE_ZIPAPPS = "--activate-zipapps"
    ENV_ALIAS = {
        "unzip": "ZIPAPPS_UNZIP",
        "unzip_exclude": "ZIPAPPS_UNZIP_EXCLUDE",
        "unzip_path": "ZIPAPPS_CACHE",
        "ignore_system_python_path": "STRICT_PYTHON_PATH",
        "python_version_slice": "PYTHON_VERSION_SLICE",
        "clear_zipapps_cache": "CLEAR_ZIPAPPS_CACHE",
        "clear_zipapps_self": "CLEAR_ZIPAPPS_SELF",
        "chmod": "UNZIP_CHMOD",
    }

    LOGGING = True

    def __init__(
        self,
        includes: str = "",
        cache_path: typing.Optional[str] = None,
        main: str = "",
        output: typing.Optional[str] = None,
        interpreter: typing.Optional[str] = None,
        compressed: bool = False,
        shell: bool = False,
        unzip: str = "",
        unzip_path: str = "",
        ignore_system_python_path=False,
        main_shell=False,
        pip_args: typing.Optional[list] = None,
        compiled: bool = False,
        build_id: str = "",
        env_paths: str = "",
        lazy_install: bool = False,
        sys_paths: str = "",
        python_version_slice: int = 2,
        ensure_pip: bool = False,
        layer_mode: bool = False,
        layer_mode_prefix: str = "python",
        clear_zipapps_cache: bool = False,
        unzip_exclude: str = "",
        chmod: str = "",
        clear_zipapps_self: bool = False,
        rm_patterns: str = "*.dist-info,__pycache__",
    ):
        """Zip your code.

        :param includes: The given paths will be copied to `cache_path` while packaging, which can be used while running. The path strings will be splited by ",". such as `my_package_dir,my_module.py,my_config.json`, often used for libs not from `pypi` or some special config files, defaults to ''
        :type includes: str, optional
        :param cache_path: if not set, will use TemporaryDirectory, prefix='zipapps_', defaults to None
        :type cache_path: str, optional
        :param main: The entry point function of the application, the `valid format` is: `package.module:function` `package.module` `module:function` `package`, defaults to ''
        :type main: str, optional
        :param output: The path of the output file, defaults to None
        :type output: str, optional
        :param interpreter: The path of the Python interpreter which will be set as the `shebang line`, defaults to `None`. With shebang `/usr/bin/python3` you can run app with `./app.pyz` directly, no need for `python3 app.pyz`, defaults to None
        :type interpreter: str, optional
        :param compressed: only for python3.7+, defaults to False
        :type compressed: bool, optional
        :param shell: whether run python in subprocess, or use runpy if shell is False, defaults to False
        :type shell: bool, optional
        :param unzip: names to be unzip, using `AUTO` is a better choice, defaults to ''. Can be overwrite with environment variable `ZIPAPPS_UNZIP`
        :type unzip: str, optional
        :param unzip_path: If `unzip` arg is not null, cache files will be unzipped to the given path while running. Defaults to `zipapps_cache`, support some internal variables: `$TEMP` means `tempfile.gettempdir()`, `$HOME` means `Path.home()`, `$SELF` means `.pyz` file path, `$PID` means `os.getpid()`, `$CWD` means `Path.cwd()`, defaults to ''
        :type unzip_path: str, optional
        :param ignore_system_python_path: Ignore global PYTHONPATH, only use zipapps_cache and app.pyz, defaults to False
        :type ignore_system_python_path: bool, optional
        :param main_shell: Only for `main` is not null, call `main` with subprocess.Popen: `python -c "import a.b;a.b.c()"`. This is used for `psutil` ImportError of DLL load, defaults to False
        :type main_shell: bool, optional
        :param pip_args: args of `pip install `, defaults to None
        :type pip_args: list, optional
        :param compiled: compile py file into pyc, defaults to False
        :type compiled: bool, optional
        :param build_id: a string to skip duplicate builds, it can be the paths of files/folders which splited by ",", then the modify time will be used as build_id. If build_id contains `*`, will use `glob` function to get paths. For example, you can set requirements.txt as your build_id by `python3 -m zipapps -b requirements.txt -r requirements.txt` when you use pyz as venv, defaults to ''
        :type build_id: str, optional
        :param env_paths: Default --zipapps arg if it is not given while running. Support $TEMP/$HOME/$SELF/$PID/$CWD prefix, separated by commas, defaults to ''
        :type env_paths: str, optional
        :param lazy_install: Install packages with pip while running, which means requirements will not be install into pyz file, defaults to False
        :type lazy_install: bool, optional
        :param sys_paths: Paths be insert to sys.path[-1] while running. Support $TEMP/$HOME/$SELF/$PID/$CWD prefix, separated by commas, defaults to ''
        :type sys_paths: str, optional
        :param python_version_slice: Only work for lazy-install mode, then `pip` target folders differ according to sys.version_info[:_slice], defaults to 2, which means 3.8.3 equals to 3.8.4 for same version accuracy 3.8, defaults to 2
        :type python_version_slice: int, optional
        :param ensure_pip: Add the ensurepip package to your pyz file, works for embed-python(windows) or other python versions without `pip` installed but `lazy-install` mode is enabled.
        :type ensure_pip: bool, optional
        :param layer_mode: Layer mode for the serverless use case, __main__.py / ensure_zipapps.py / activate_zipapps.py files will not be set in this mode, which means it will skip the activative process.
        :type layer_mode: bool, optional
        :param layer_mode_prefix: Only work while --layer-mode is set, will move the files in the given prefix folder.
        :type layer_mode_prefix: str, optional
        :param clear_zipapps_cache: Clear the zipapps cache folder after running, but maybe failed for .pyd/.so files.
        :type clear_zipapps_cache: bool, optional
        :param unzip_exclude: names not to be unzip, defaults to '', should be used with unzip. Can be overwrite with environment variable `ZIPAPPS_UNZIP_EXCLUDE`
        :type unzip_exclude: str, optional
        :param chmod: os.chmod(int(chmod, 8)) for unzip files with `--chmod=777`, unix-like system only
        :type chmod: str, optional
        :param clear_zipapps_self: Clear the zipapps pyz file after running.
        :type clear_zipapps_self: bool, optional
        :param rm_patterns: Delete useless files or folders, splited by "," and defaults to `*.dist-info,__pycache__`. Recursively glob: **/*.pyc
        :type rm_patterns: str
        """
        self.includes = includes
        self.cache_path = cache_path
        self.main = main
        self.output = output or self.DEFAULT_OUTPUT_PATH
        self._output_path = Path(self.output)
        self.interpreter = interpreter
        self.compressed = compressed
        self.shell = shell
        self.unzip = unzip
        self.unzip_exclude = unzip_exclude
        self.unzip_path = unzip_path
        self.ignore_system_python_path = ignore_system_python_path
        self.main_shell = main_shell
        self.pip_args = pip_args
        self.compiled = compiled
        self.build_id = build_id
        self.env_paths = env_paths
        self.lazy_install = lazy_install
        self.sys_paths = sys_paths
        self.python_version_slice = python_version_slice
        self.ensure_pip = ensure_pip
        self.layer_mode = layer_mode
        self.layer_mode_prefix = layer_mode_prefix
        self.clear_zipapps_cache = clear_zipapps_cache
        self.clear_zipapps_self = clear_zipapps_self
        self.chmod = chmod
        self.rm_patterns = rm_patterns

        self._tmp_dir: typing.Optional[tempfile.TemporaryDirectory] = None
        self._build_success = False
        self._is_greater_than_python_37 = (
            sys.version_info.minor >= 7 and sys.version_info.major >= 3
        )

    @property
    def kwargs(self):
        return dict(
            includes=self.includes,
            cache_path=str(self.cache_path or ""),
            main=self.main,
            output=self.output,
            interpreter=self.interpreter,
            compressed=self.compressed,
            shell=self.shell,
            unzip=self.unzip,
            unzip_path=self.unzip_path,
            ignore_system_python_path=self.ignore_system_python_path,
            main_shell=self.main_shell,
            pip_args=self.pip_args,
            compiled=self.compiled,
            build_id=self.build_id,
            env_paths=self.env_paths,
            lazy_install=self.lazy_install,
            sys_paths=self.sys_paths,
            python_version_slice=self.python_version_slice,
            ensure_pip=self.ensure_pip,
            layer_mode=self.layer_mode,
            layer_mode_prefix=self.layer_mode_prefix,
            clear_zipapps_cache=self.clear_zipapps_cache,
            unzip_exclude=self.unzip_exclude,
            chmod=self.chmod,
            clear_zipapps_self=self.clear_zipapps_self,
        )

    def ensure_args(self):
        if not self.unzip:
            if self.unzip_exclude:
                self._log(
                    "[WARN]: The arg `unzip_exclude` should not be with `unzip` but `unzip` is null."
                )
            if self.compiled:
                self._log(
                    "[WARN]: The arg `compiled` should not be True while `unzip` is null, because .pyc files of __pycache__ folder may not work in zip file."
                )
            if self.lazy_install:
                self._log(
                    '[WARN]: the `unzip` arg has been changed to "*" while `lazy_install` is True.'
                )
                self.unzip = "*"
        if self.cache_path:
            self._cache_path = Path(self.cache_path)
        else:
            self._tmp_dir = tempfile.TemporaryDirectory(prefix="zipapps_")
            self._cache_path = Path(self._tmp_dir.name)
        if not self.unzip_path:
            if self.lazy_install:
                self._log(
                    f"[WARN]: the arg `unzip_path` has been changed to `SELF/{self.DEFAULT_UNZIP_CACHE_PATH}` while `lazy_install` is True and `unzip_path` is null."
                )
                self.unzip_path = f"SELF/{ZipApp.DEFAULT_UNZIP_CACHE_PATH}"
            else:
                self._log(
                    f"[INFO]: the arg `unzip_path` has been changed to `{self.DEFAULT_UNZIP_CACHE_PATH}` by default."
                )
                self.unzip_path = self.DEFAULT_UNZIP_CACHE_PATH
        self.build_id_name = self.get_build_id_name()
        self._log(
            f"[INFO]: output path is `{self._output_path}`, you can reset it with the arg `output`."
        )

    def prepare_ensure_pip(self):
        if self.ensure_pip:
            import ensurepip

            ensurepip_dir_path = Path(ensurepip.__file__).parent
            shutil.copytree(
                str(ensurepip_dir_path.absolute()),
                self._cache_path / ensurepip_dir_path.name,
            )

    def build(self):
        self._log(
            f'[INFO]: {"=" * 10} Start building `{self._output_path}` with zipapps version <{__version__}> {"=" * 10}'
        )
        self.ensure_args()
        if self.build_exists():
            return self._output_path
        self.prepare_includes()
        self.prepare_ensure_pip()
        self.prepare_pip()
        if not self.layer_mode:
            self.prepare_entry_point()
        if self.build_id_name:
            # make build_id file
            (self._cache_path / self.build_id_name).touch()
        if self.compiled:
            compileall.compile_dir(self._cache_path, **ZipApp.COMPILE_KWARGS)
        self.clean_pip_pycache()
        if self.layer_mode:
            self.create_archive_layer()
        else:
            self.create_archive()
        self._build_success = True
        return self._output_path

    def create_archive_layer(self):
        if self.compressed:
            compression = ZIP_STORED
            compresslevel = 9
        else:
            compression = ZIP_DEFLATED
            compresslevel = 0
        _kwargs = dict(mode="w", compression=compression)
        if self._is_greater_than_python_37:
            _kwargs["compresslevel"] = compresslevel
        with ZipFile(str(self._output_path), **_kwargs) as zf:
            for f in self._cache_path.glob("**/*"):
                zf.write(f, str(f.relative_to(self._cache_path)))

    def create_archive(self):
        if self._is_greater_than_python_37:
            zipapp.create_archive(
                source=self._cache_path,
                target=str(self._output_path.absolute()),
                interpreter=self.interpreter,
                compressed=self.compressed,
            )
        elif self.compressed:
            raise RuntimeError("The arg `compressed` only support python3.7+")
        else:
            zipapp.create_archive(
                source=self._cache_path,
                target=str(self._output_path.absolute()),
                interpreter=self.interpreter,
            )

    def prepare_entry_point(self):
        # reset unzip_names
        unzip_names = set(self.unzip.split(",")) if self.unzip else set()
        warning_names: typing.Dict[str, dict] = {}
        for path in self._cache_path.iterdir():
            _name_not_included = path.name not in unzip_names
            if path.is_dir():
                pyd_counts = len(list(path.glob("**/*.pyd")))
                so_counts = len(list(path.glob("**/*.so")))
                if (pyd_counts or so_counts) and _name_not_included:
                    # warn which libs need to be unzipped
                    if pyd_counts:
                        warning_names.setdefault(path.name, {})[".pyd"] = pyd_counts
                    if so_counts:
                        warning_names.setdefault(path.name, {})[".so"] = so_counts
            elif path.is_file() and path.suffix in (".pyd", ".so"):
                if _name_not_included and path.stem not in unzip_names:
                    warning_names.setdefault(path.name, {})[path.suffix] = 1
        # remove the special keys from unzip_names
        auto_unzip_keys = ZipApp.AUTO_FIX_UNZIP_KEYS & unzip_names
        unzip_names -= auto_unzip_keys
        if warning_names:
            if self.clear_zipapps_cache:
                msg = f"[WARN]: clear_zipapps_cache is True but .pyd/.so files were found {warning_names}"
                self._log(msg)
            if auto_unzip_keys:
                unzip_names |= warning_names.keys()
            else:
                _fix_unzip_names = ",".join(warning_names.keys())
                msg = f"[WARN]: .pyd/.so files may be imported incorrectly, set `--unzip={_fix_unzip_names}` or `--unzip=AUTO` or `--unzip=*` to fix it. {warning_names}"
                self._log(msg)
        new_unzip = ",".join(unzip_names)
        self.unzip = new_unzip
        if self.unzip:
            self._log(
                f"[INFO]: these names will be unzipped while running: {self.unzip}"
            )
        self.prepare_active_zipapps()

    def prepare_active_zipapps(self):
        output_name = Path(self._output_path).stem
        if not re.match(r"^[0-9a-zA-Z_]+$", output_name):
            raise ValueError("The name of `output` should match regex: ^[0-9a-zA-Z_]+$")

        def make_runner():
            if self.main:
                pattern = r"^\w+(\.\w+)?(:\w+)?$"
                if re.match(pattern, self.main):
                    module, _, function = self.main.partition(":")
                    if module:
                        # main may be: 'module.py:main' or 'module.submodule:main'
                        # replace module.py to module
                        module_path = self._cache_path / module
                        if module_path.is_file():
                            module = module_path.stem
                        runner = f"import {module}"
                        if function:
                            runner += f"; {module}.{function}()"
                        self._log(
                            f"[INFO]: -m: matches re.match(r'{pattern}', self.main), add as `{runner}`."
                        )
                        return runner
                else:
                    self._log(
                        f"[INFO]: -m: not matches re.match(r'{pattern}', self.main), add as raw code `{self.main}`."
                    )
                    return self.main
            return ""

        kwargs = {
            "ts": self.setup_timestamp_file(),
            "shell": self.shell,
            "main_shell": self.main_shell,
            "unzip": repr(self.unzip),
            "unzip_exclude": repr(self.unzip_exclude),
            "output_name": output_name,
            "unzip_path": repr(self.unzip_path),
            "ignore_system_python_path": self.ignore_system_python_path,
            "has_main": bool(self.main),
            "run_main": make_runner(),
            "HANDLE_OTHER_ENVS_FLAG": self.HANDLE_OTHER_ENVS_FLAG,
            "env_paths": repr(self.env_paths),
            "LAZY_PIP_DIR_NAME": repr(self.LAZY_PIP_DIR_NAME),
            "pip_args_repr": repr(self.pip_args),
            "sys_paths": repr(self.sys_paths),
            "python_version_slice": repr(self.python_version_slice),
            "pip_args_md5": self.pip_args_md5,
            "clear_zipapps_cache": repr(self.clear_zipapps_cache),
            "HANDLE_ACTIVATE_ZIPAPPS": self.HANDLE_ACTIVATE_ZIPAPPS,
            "chmod": repr(self.chmod),
            "clear_zipapps_self": repr(self.clear_zipapps_self),
        }
        for k, v in self.ENV_ALIAS.items():
            kwargs[f"{k}_env"] = repr(v)
        code = get_data(__name__, "entry_point.py.template").decode("u8")
        (self._cache_path / "__main__.py").write_text(code.format(**kwargs))

        code = get_data(__name__, "ensure_zipapps.py.template").decode("u8")
        (self._cache_path / "ensure_zipapps.py").write_text(code.format(**kwargs))

        code = get_data(__name__, "activate_zipapps.py").decode("u8")
        (self._cache_path / "activate_zipapps.py").write_text(code)
        code += "\n\nactivate()"

        if output_name != "zipapps":
            (self._cache_path / f"ensure_{output_name}.py").write_text(code)
        (self._cache_path / f"ensure_zipapps_{output_name}.py").write_text(code)
        (self._cache_path / "zipapps_config.json").write_text(json.dumps(self.kwargs))

    def setup_timestamp_file(
        self,
    ):
        ts = str(int(time.time() * 10000000))
        (self._cache_path / ("_zip_time_%s" % ts)).touch()
        return ts

    @staticmethod
    def get_md5(value: typing.Any):
        if not isinstance(value, bytes):
            value = str(value).encode("utf-8")
        return md5(value).hexdigest()

    def prepare_pip(self):
        self.pip_args_md5 = ""
        if self.pip_args:
            if "-t" in self.pip_args or "--target" in self.pip_args:
                raise RuntimeError(
                    "`-t` / `--target` arg can be set with `--cache-path`/`cache_path` to rewrite the zipapps cache path."
                )
            if self.lazy_install:
                # copy files to cache folder
                _temp_pip_path = self._cache_path / self.LAZY_PIP_DIR_NAME
                _temp_pip_path.mkdir(parents=True, exist_ok=True)
                _md5_str = self.get_md5(self.pip_args)
                # overwrite path args to new path, such as requirements.txt or xxx.whl
                for index, arg in enumerate(self.pip_args):
                    path = Path(arg)
                    if path.is_file():
                        _md5_str += self.get_md5(path.read_bytes())
                        new_path = _temp_pip_path / path.name
                        shutil.copyfile(path, new_path)
                        _r_path = Path(self.LAZY_PIP_DIR_NAME) / path.name
                        self.pip_args[index] = _r_path.as_posix()
                self.pip_args_md5 = self.get_md5(_md5_str.encode("utf-8"))
                self._log(
                    f"[INFO]: pip_args_md5 has been generated: {self.pip_args_md5}"
                )
            else:
                self.pip_install()

    @classmethod
    def _rm_with_patterns(
        cls,
        target_dir: Path,
        patterns=("*.dist-info", "__pycache__"),
    ):
        target_dir = Path(target_dir)
        for pattern in patterns:
            if pattern:
                for path in target_dir.glob(pattern):
                    if path.is_dir():
                        shutil.rmtree(path, ignore_errors=True)
                    else:
                        try:
                            path.unlink()
                        except FileNotFoundError:
                            pass

    @classmethod
    def _pip_install(cls, target_dir: Path, pip_args: list):
        target_dir = Path(target_dir)
        _pip_args = [
            "install",
            "--target",
            target_dir.absolute().as_posix(),
        ] + pip_args
        pip_main = get_pip_main()
        result = pip_main(_pip_args)
        if result != 0:
            raise RuntimeError("pip install failed: return code=%s" % result)

    def clean_pip_pycache(self):
        if self.layer_mode:
            target_dir = self._cache_path / self.layer_mode_prefix
        else:
            target_dir = self._cache_path
        return self._rm_with_patterns(target_dir, patterns=self.rm_patterns.split(","))

    def pip_install(self):
        if self.layer_mode:
            _target_dir = self._cache_path.absolute() / self.layer_mode_prefix
        else:
            _target_dir = self._cache_path
        return self._pip_install(target_dir=_target_dir, pip_args=self.pip_args)

    def prepare_includes(self):
        if not self.includes:
            return
        if self.layer_mode:
            _target_dir = self._cache_path.absolute() / self.layer_mode_prefix
        else:
            _target_dir = self._cache_path.absolute()
        _target_dir.mkdir(parents=True, exist_ok=True)
        for _include_path in self.includes.split(self.PATH_SPLIT_TAG):
            include_path = Path(_include_path)
            if include_path.is_dir():
                shutil.copytree(include_path, _target_dir / include_path.name)
            elif include_path.is_file():
                shutil.copyfile(include_path, _target_dir / include_path.name)
            else:
                raise RuntimeError("%s is not exist" % include_path.absolute())

    def build_exists(self):
        if self.build_id_name and self._output_path.is_file():
            try:
                with ZipFile(self._output_path, "r") as zf:
                    for member in zf.infolist():
                        if member.filename == self.build_id_name:
                            return True
            except BadZipFile:
                pass
        return False

    def get_build_id_name(self):
        if not self.build_id:
            return ""
        build_id_str = ""
        if "*" in self.build_id:
            paths = glob(self.build_id)
        else:
            paths = self.build_id.split(",")
        for p in paths:
            try:
                path = Path(p)
                build_id_str += str(path.stat().st_mtime)
            except FileNotFoundError:
                pass
        build_id_str = build_id_str or str(self.build_id)
        md5_id = self.get_md5(build_id_str.encode("utf-8"))
        return f"_build_id_{md5_id}"

    @classmethod
    def create_app(
        cls,
        includes: str = "",
        cache_path: typing.Optional[str] = None,
        main: str = "",
        output: typing.Optional[str] = None,
        interpreter: typing.Optional[str] = None,
        compressed: bool = False,
        shell: bool = False,
        unzip: str = "",
        unzip_path: str = "",
        ignore_system_python_path=False,
        main_shell=False,
        pip_args: typing.Optional[list] = None,
        compiled: bool = False,
        build_id: str = "",
        env_paths: str = "",
        lazy_install: bool = False,
        sys_paths: str = "",
        python_version_slice: int = 2,
        ensure_pip: bool = False,
        layer_mode: bool = False,
        layer_mode_prefix: str = "python",
        clear_zipapps_cache: bool = False,
        unzip_exclude: str = "",
        chmod: str = "",
        clear_zipapps_self: bool = False,
        rm_patterns: str = "*.dist-info,__pycache__",
    ):
        app = cls(
            includes=includes,
            cache_path=cache_path,
            main=main,
            output=output,
            interpreter=interpreter,
            compressed=compressed,
            shell=shell,
            unzip=unzip,
            unzip_path=unzip_path,
            ignore_system_python_path=ignore_system_python_path,
            main_shell=main_shell,
            pip_args=pip_args,
            compiled=compiled,
            build_id=build_id,
            env_paths=env_paths,
            lazy_install=lazy_install,
            sys_paths=sys_paths,
            python_version_slice=python_version_slice,
            ensure_pip=ensure_pip,
            layer_mode=layer_mode,
            layer_mode_prefix=layer_mode_prefix,
            clear_zipapps_cache=clear_zipapps_cache,
            unzip_exclude=unzip_exclude,
            chmod=chmod,
            clear_zipapps_self=clear_zipapps_self,
            rm_patterns=rm_patterns,
        )
        return app.build()

    @classmethod
    def _log(cls, text):
        if cls.LOGGING:
            sys.stderr.write(f'{time.strftime("%Y-%m-%d %H:%M:%S")} | {text}\n')

    def __del__(self):
        if self._tmp_dir:
            self._tmp_dir.cleanup()
            self._log(f"[INFO]: Temp cache has been cleaned. ({self._tmp_dir!r})")
        if self._build_success:
            self._log(
                f'[INFO]: {"=" * 10} Successfully built `{self._output_path}` {"=" * 10}'
            )
        else:
            self._log(f'[ERROR]: {"=" * 10} Build failed {"=" * 10}')


def pip_install_target(
    target: Path,
    pip_args: list,
    rm_patterns: str = "*.dist-info,__pycache__",
    force=False,
    clear_dir=False,
    sys_path: typing.Optional[int] = None,
):
    """
    Installs target dependencies using pip and cleans up the installed files afterwards.

    Args:
        target (Path): The target directory for dependency installation.
        pip_args (list): List of arguments for the pip install command.
        rm_patterns (str, optional): Patterns of files to remove after installation, defaults to "*.dist-info,__pycache__".
        force (bool, optional): Flag to force reinstallation, defaults to False.
        clear_dir (bool, optional): will rmtree the target directory while clear_dir=True.
        sys_path (typing.Optional[int], optional): If provided, inserts the target directory at the specified position in sys.path.

    Returns:
        bool: Returns True if the installation is successful.
    """
    target = Path(target)
    md5_path = target / ZipApp.get_md5(pip_args)
    if not force and md5_path.exists():
        return False
    if clear_dir and target.is_dir():
        shutil.rmtree(target.as_posix(), ignore_errors=True)
    ZipApp._pip_install(target_dir=target, pip_args=pip_args)
    if rm_patterns:
        ZipApp._rm_with_patterns(target, patterns=rm_patterns.split(","))

    md5_path.touch()
    if isinstance(sys_path, int):
        sys.path.insert(sys_path, target.absolute().as_posix())
    return True


create_app = ZipApp.create_app



import asyncio
import inspect
import json
import time
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from itertools import chain
from threading import Lock, Semaphore
from typing import Callable, Coroutine, Dict, Optional, OrderedDict, Set, Union
from weakref import WeakSet

__all__ = ["lru_cache_ttl", "threads", "bg_task", "NamedLock", "FuncSchema"]


def lru_cache_ttl(
    maxsize: int,
    ttl: Optional[Union[int, float]] = None,
    controls=False,
    auto_clear=True,
    timer=time.time,
):
    """A Least Recently Used (LRU) cache with a Time To Live (TTL) feature.

    Args:
        maxsize (int): maxsize of cache
        ttl (Optional[Union[int, float]], optional): time to live. Defaults to None.
        controls (bool, optional): set cache/ttl_clean attributes. Defaults to False.
        auto_clear (bool, optional): clear dead cache automatically. Defaults to True.
        timer (callable, optional): Defaults to time.time.

    Returns:
        callable: decorator function

    >>> import time
    >>> # test ttl
    >>> values = [1, 2]
    >>> @lru_cache_ttl(1, 0.1)
    ... def func1(i):
    ...     return values.pop(0)
    >>> [func1(1), func1(1), time.sleep(0.11), func1(1)]
    [1, 1, None, 2]
    >>> # test maxsize
    >>> values = [1, 2, 3]
    >>> func = lambda i: values.pop(0)
    >>> func1 = lru_cache_ttl(2)(func)
    >>> [func1(i) for i in [1, 1, 1, 2, 2, 2, 3, 3, 3]]
    [1, 1, 1, 2, 2, 2, 3, 3, 3]
    >>> # test auto_clear=True, with controls
    >>> values = [1, 2, 3, 4]
    >>> func = lambda i: values.pop(0)
    >>> func1 = lru_cache_ttl(5, 0.1, controls=True, auto_clear=True)(func)
    >>> [func1(1), func1(2), func1(3)]
    [1, 2, 3]
    >>> time.sleep(0.11)
    >>> func1(3)
    4
    >>> len(func1.cache)
    1
    >>> # test auto_clear=False
    >>> values = [1, 2, 3, 4]
    >>> @lru_cache_ttl(5, 0.1, controls=True, auto_clear=False)
    ... def func1(i):
    ...     return values.pop(0)
    >>> [func1(1), func1(2), func1(3)]
    [1, 2, 3]
    >>> time.sleep(0.11)
    >>> func1(3)
    4
    >>> len(func1.cache)
    3
    """
    cache: OrderedDict = OrderedDict()
    move_to_end = cache.move_to_end
    popitem = cache.popitem
    next_clear_ts = timer()

    def ttl_clean(expire):
        for k, v in tuple(cache.items()):
            if v[1] <= expire:
                cache.pop(k, None)

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            nonlocal next_clear_ts
            now = timer()
            if ttl is None:
                expire_cleared = True
            else:
                expire_cleared = auto_clear and now > next_clear_ts
                if expire_cleared:
                    ttl_clean(now)
                    next_clear_ts = now + ttl
            # key = _make_key(args, kwargs, False)
            key = hash(tuple(chain(args, kwargs.items())))
            if key in cache:
                if expire_cleared is True or now < cache[key][1]:
                    # 1. key in cache; 2. cache is valid
                    # set newest, return result
                    move_to_end(key)
                    return cache[key][0]
            # call function
            result = func(*args, **kwargs)
            # ensure size
            while len(cache) >= maxsize:
                popitem(last=False)
            if ttl is None:
                # no need save expired ts
                cache[key] = (result,)
            else:
                # setitem(key, (result, now + ttl))
                cache[key] = (result, timer() + ttl)
            # {key: (result, expired_ts)}
            return result

        if controls:
            setattr(wrapped, "cache", cache)
            setattr(wrapped, "ttl_clean", ttl_clean)
        return wrapped

    return decorator


def threads(n: Optional[int] = None, executor_class=None, **kws):
    """Quickly convert synchronous functions to be concurrency-able. (similar to madisonmay/Tomorrow)

    >>> @threads(10)
    ... def test(i):
    ...     time.sleep(i)
    ...     return i
    >>> start = time.time()
    >>> tasks = [test(i) for i in [0.1] * 5]
    >>> len(test.pool._threads)
    5
    >>> len(test.tasks)
    5
    >>> for i in tasks:
    ...     i.result() if hasattr(i, 'result') else i
    0.1
    0.1
    0.1
    0.1
    0.1
    >>> time.time() - start < 0.2
    True
    >>> len(test.pool._threads)
    5
    >>> len(test.tasks)
    0
    """
    pool = (executor_class or ThreadPoolExecutor)(max_workers=n, **kws)
    tasks: WeakSet = WeakSet()

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            future = pool.submit(func, *args, **kwargs)
            future.add_done_callback(lambda task: tasks.discard(task))
            tasks.add(future)
            return future

        setattr(wrapped, "pool", pool)
        setattr(wrapped, "tasks", tasks)
        return wrapped

    return decorator


_bg_tasks: Set[asyncio.Task] = set()


def bg_task(coro: Coroutine) -> asyncio.Task:
    """Avoid asyncio free-flying tasks, better to use the new asyncio.TaskGroup to avoid this in 3.11+. https://github.com/python/cpython/issues/91887

    Args:
        coro (Coroutine)

    Returns:
        _type_: Task

    """
    task = asyncio.create_task(coro)
    _bg_tasks.add(task)
    task.add_done_callback(_bg_tasks.discard)
    return task


LockType = Union[Lock, Semaphore, asyncio.Lock, asyncio.Semaphore]


class NamedLock:
    """Reusable named locks, support for timeouts, support for multiple concurrent locks.

            ```python

    def test_named_lock():
        def test_sync():
            import time
            from concurrent.futures import ThreadPoolExecutor
            from threading import Lock, Semaphore

            def _test1():
                with NamedLock("_test1", Lock, timeout=0.05) as lock:
                    time.sleep(0.2)
                    return bool(lock)

            with ThreadPoolExecutor(10) as pool:
                tasks = [pool.submit(_test1) for _ in range(3)]
                result = [i.result() for i in tasks]
                assert result == [True, False, False], result
            assert len(NamedLock._SYNC_CACHE) == 1
            NamedLock.clear_unlocked()
            assert len(NamedLock._SYNC_CACHE) == 0

            def _test2():
                with NamedLock("_test2", lambda: Semaphore(2), timeout=0.05) as lock:
                    time.sleep(0.2)
                    return bool(lock)

            with ThreadPoolExecutor(10) as pool:
                tasks = [pool.submit(_test2) for _ in range(3)]
                result = [i.result() for i in tasks]
                assert result == [True, True, False], result

        def test_async():
            import asyncio

            async def main():
                async def _test1():
                    async with NamedLock("_test1", asyncio.Lock, timeout=0.05) as lock:
                        await asyncio.sleep(0.2)
                        return bool(lock)

                tasks = [asyncio.create_task(_test1()) for _ in range(3)]
                result = [await i for i in tasks]
                assert result == [True, False, False], result
                assert len(NamedLock._ASYNC_CACHE) == 1
                NamedLock.clear_unlocked()
                assert len(NamedLock._ASYNC_CACHE) == 0

                async def _test2():
                    async with NamedLock(
                        "_test2", lambda: asyncio.Semaphore(2), timeout=0.05
                    ) as lock:
                        await asyncio.sleep(0.2)
                        return bool(lock)

                tasks = [asyncio.create_task(_test2()) for _ in range(3)]
                result = [await i for i in tasks]
                assert result == [True, True, False], result

            asyncio.get_event_loop().run_until_complete(main())

        test_sync()
        test_async()

            ```
    """

    _SYNC_CACHE: Dict[str, LockType] = {}
    _ASYNC_CACHE: Dict[str, LockType] = {}

    def __init__(self, name: str, default_factory: Callable, timeout=None):
        self.name = name
        self.default_factory = default_factory
        self.timeout = timeout
        self.lock = None

    @classmethod
    def clear_unlocked(cls):
        for cache in [cls._SYNC_CACHE, cls._ASYNC_CACHE]:
            for name, lock in list(cache.items()):
                if hasattr(lock, "locked") and not lock.locked():
                    cache.pop(name, None)
                elif isinstance(lock, Semaphore) and (
                    lock._value == 0 or not lock._cond._lock.locked()
                ):
                    cache.pop(name, None)

    def __enter__(self):
        if self.name in self._SYNC_CACHE:
            lock = self._SYNC_CACHE[self.name]
        else:
            lock = self.default_factory()
            self._SYNC_CACHE[self.name] = lock
        if lock.acquire(timeout=self.timeout):
            self.lock = lock
            return self
        else:
            return None

    def __exit__(self, *_):
        if self.lock:
            self.lock.release()

    async def __aenter__(self):
        if self.name in self._ASYNC_CACHE:
            lock = self._ASYNC_CACHE[self.name]
        else:
            lock = self.default_factory()
            self._ASYNC_CACHE[self.name] = lock
        try:
            await asyncio.wait_for(lock.acquire(), timeout=self.timeout)
            self.lock = lock
            return self
        except asyncio.TimeoutError:
            return None

    async def __aexit__(self, *_):
        if self.lock:
            self.lock.release()


class FuncSchema:
    """Parse the parameters and types required by a function into a dictionary, and convert an incoming parameter into the appropriate type.

    >>> def test(a, b: str, /, c=1, *, d=["d"], e=0.1, f={"f"}, g=(1, 2), h=True, i={1}, **kws):
    ...     return
    >>> FuncSchema.parse(test)
    {'b': {'type': <class 'str'>, 'default': <class 'inspect._empty'>}, 'c': {'type': <class 'int'>, 'default': 1}, 'd': {'type': <class 'list'>, 'default': ['d']}, 'e': {'type': <class 'float'>, 'default': 0.1}, 'f': {'type': <class 'set'>, 'default': {'f'}}, 'g': {'type': <class 'tuple'>, 'default': (1, 2)}, 'h': {'type': <class 'bool'>, 'default': True}, 'i': {'type': <class 'set'>, 'default': {1}}}
    >>> FuncSchema.convert("1", int)
    1
    >>> FuncSchema.convert("1", str)
    '1'
    >>> FuncSchema.convert("1", float)
    1.0
    >>> FuncSchema.convert(0, bool)
    False
    >>> FuncSchema.convert('1', bool)
    True
    >>> FuncSchema.convert('[[1, 1]]', dict)
    {1: 1}
    >>> FuncSchema.convert('[1, 1]', set)
    {1}
    >>> FuncSchema.convert('[1, 1]', tuple)
    (1, 1)
    """

    ALLOW_TYPES = {int, float, str, tuple, list, set, dict, bool}
    JSON_TYPES = {tuple, set, dict, bool}

    @classmethod
    def parse(cls, function: Callable):
        sig = inspect.signature(function)
        result = {}
        for param in sig.parameters.values():
            if param.annotation is param.empty:
                if param.default is param.empty:
                    continue
                tp = type(param.default)
            else:
                tp = param.annotation
            if tp in cls.ALLOW_TYPES:
                result[param.name] = {"type": tp, "default": param.default}
        return result

    @classmethod
    def convert(cls, obj, target_type):
        if isinstance(obj, str) and target_type in cls.JSON_TYPES:
            return target_type(json.loads(obj))
        else:
            return target_type(obj)


def test_bg_task():
    async def _test_bg_task():
        async def coro():
            return True

        task = bg_task(coro())
        assert await task is True
        result = (task.done(), len(_bg_tasks))
        assert result == (True, 0), result

    asyncio.get_event_loop().run_until_complete(_test_bg_task())


def test_named_lock():
    def test_sync():
        import time
        from concurrent.futures import ThreadPoolExecutor
        from threading import Lock, Semaphore

        def _test1():
            with NamedLock("_test1", Lock, timeout=0.05) as lock:
                time.sleep(0.2)
                return bool(lock)

        with ThreadPoolExecutor(10) as pool:
            tasks = [pool.submit(_test1) for _ in range(3)]
            result = [i.result() for i in tasks]
            assert result == [True, False, False], result
        assert len(NamedLock._SYNC_CACHE) == 1
        NamedLock.clear_unlocked()
        assert len(NamedLock._SYNC_CACHE) == 0

        def _test2():
            with NamedLock("_test2", lambda: Semaphore(2), timeout=0.05) as lock:
                time.sleep(0.2)
                return bool(lock)

        with ThreadPoolExecutor(10) as pool:
            tasks = [pool.submit(_test2) for _ in range(3)]
            result = [i.result() for i in tasks]
            assert result == [True, True, False], result

    def test_async():
        import asyncio

        async def main():
            async def _test1():
                async with NamedLock("_test1", asyncio.Lock, timeout=0.05) as lock:
                    await asyncio.sleep(0.2)
                    return bool(lock)

            tasks = [asyncio.create_task(_test1()) for _ in range(3)]
            result = [await i for i in tasks]
            assert result == [True, False, False], result
            assert len(NamedLock._ASYNC_CACHE) == 1
            NamedLock.clear_unlocked()
            assert len(NamedLock._ASYNC_CACHE) == 0

            async def _test2():
                async with NamedLock(
                    "_test2", lambda: asyncio.Semaphore(2), timeout=0.05
                ) as lock:
                    await asyncio.sleep(0.2)
                    return bool(lock)

            tasks = [asyncio.create_task(_test2()) for _ in range(3)]
            result = [await i for i in tasks]
            assert result == [True, True, False], result

        asyncio.get_event_loop().run_until_complete(main())

    test_sync()
    test_async()


def test():
    test_bg_task()
    test_named_lock()


if __name__ == "__main__":
    __name__ = "morebuiltins.functools"
    test()
    import doctest

    doctest.testmod()



import re
from datetime import datetime, timedelta
from typing import Callable, Dict, List, Optional, Tuple, Union


# NotImplemented yet
class ScheduleTimer:
    """The ScheduleTimer class facilitates the creation and evaluation of datetime patterns for scheduling purposes.

    It includes mechanisms to parse patterns involving logical operations (AND, OR) and comparison checks (equality, inequality, arithmetic, and custom range checks).

    Comparison Operators:

        Equality (= or ==): Tests for equality between datetime parts.
            Example: "hour=12" checks if it's exactly 12 o'clock.
        Inequality (!=): Ensures inequality between datetime parts.
            Example: "minute!=30" for minutes not being 30.
        Less Than (<): Requires the left datetime part to be less than the right.
            Example: "day<15" for days before the 15th.
        Less Than or Equal (<=): Allows the left datetime part to be less or equal to the right.
            Example: "month<=6" covers January to June.
        Greater Than (>): Ensures the left datetime part is greater than the right.
            Example: "hour>18" for evenings after 6 PM.
        Greater Than or Equal (>=): Allows the left datetime part to be greater or equal to the right.
            Example: "weekday>=MO" starts from Monday.
        Division Modulo (/): Divisibility check for the left digit by the right digit in datetime format.
            Example: "minute/15" checks for quarter hours.
        Range Inclusion (@): Confirms if the left time falls within any defined ranges in the right string.
            Example: "hour@9-11,13-15" for office hours.

    Logical Operators:
        AND (&): Both conditions must hold true.
            Example: "hour=12&minute=30" for exactly 12:30 PM.
        OR (; or |): At least one of the conditions must be true.
            Example: "hour=12;hour=18" for noon or 6 PM.
        Negation (!): Inverts the truth of the following condition.
            Example: "!hour=12" excludes noon.

    Typical Pattern Examples:

        Daily at Midnight
            Pattern: "hour=00&minute=00&second=00"
        Weekdays Except Fridays, Every 3 Hours Starting at 9 AM
            Pattern: "weekday>=MO&weekday<=TH&hour>=9;hour=9+3/3"
        First Monday of Every Month at 10 AM
            Pattern: "day=1&weekday=MO&hour=10&minute=00"
        Every 15 Minutes During Business Hours
            Pattern: "hour>=9&hour<=17&minute/15"
        Alternating Fridays, 2 PM to 4 PM
            Pattern: "weekday=FRI&hour>=14&hour<=16&!(weekday@2,4,6,8,10,12)"
    """

    prefix_not = "!"
    split_or = re.compile(";|\|")
    split_and = re.compile("&")
    split_calc = re.compile("(=+|!=|<=?|>=?|@|/)")
    magic_methods = {
        "=": "__eq__",
        "==": "__eq__",
        "!=": "__ne__",
        "<": "__lt__",
        "<=": "__le__",
        ">": "__gt__",
        ">=": "__ge__",
    }
    custom_methods = {"/": "divide", "@": "between"}
    step_choices = [
        ("%S", timedelta(seconds=1)),
        ("%M", timedelta(minutes=1)),
        ("%H", timedelta(hours=1)),
    ]

    def __init__(self):
        raise NotImplementedError("Performance issue needs to be optimized")

    @staticmethod
    def divide(a: str, b: str):
        if a.isdigit() and b.isdigit():
            return int(a) % int(b) == 0
        else:
            return False

    @staticmethod
    def between(a: str, b: str):
        # many ranges
        range_regex = re.compile("^(\d+)-(\d+)$")
        for time_range in b.split(","):
            if a == time_range:
                return True
            else:
                match = range_regex.match(time_range)
                if match and int(a) >= int(match[1]) and int(a) <= int(match[2]):
                    return True
        return False

    @classmethod
    def get_step(cls, pattern: str):
        for step, delta in cls.step_choices:
            if step in pattern:
                return delta
        return timedelta(days=1)

    @classmethod
    def sort_pattern(cls, pattern: str):
        steps: Dict[str, timedelta] = {}
        for or_part in cls.split_or.split(pattern):
            and_parts = []
            for and_part in cls.split_and.split(or_part):
                steps[and_part] = cls.get_step(and_part)
                and_parts.append(and_part)
            and_parts.sort(key=lambda x: steps[x], reverse=True)
        return pattern, steps

    @classmethod
    def get_next_datetimes(
        cls,
        pattern: str,
        count: int = 1,
        start_date: Optional[datetime] = None,
        max_tries: int = 1000000,
        callback: Callable = lambda x: x,
    ) -> List[Union[datetime, str]]:
        if start_date is None:
            start_date = datetime.now().replace(microsecond=0)
        result: List[Union[datetime, str]] = []
        pattern_list, steps = cls.sort_pattern(pattern)
        min_step = min(steps.values())
        min_step = timedelta(seconds=1)
        for _ in range(max_tries):
            ok, miss_pattern = cls.is_at(pattern_list, start_date)
            if ok:
                item = callback(start_date)
                if not result or item != result[-1]:
                    result.append(item)
                    if len(result) == count:
                        break
                step = min_step
            else:
                step = steps[miss_pattern]
            start_date += step
        return result

    @classmethod
    def is_at(
        cls, pattern: Union[list, str], date: Optional[datetime] = None
    ) -> Tuple[bool, str]:
        if date is None:
            date = datetime.now()
        if isinstance(pattern, str):
            pattern_list, _ = cls.sort_pattern(pattern)
        else:
            pattern_list = pattern
        miss_pattern: str = ""
        ok = False
        for or_pattern in cls.split_or.split(pattern_list):
            for and_pattern in cls.split_and.split(or_pattern):
                if and_pattern[0] == cls.prefix_not:
                    and_pattern = and_pattern[1:]
                    be_not = True
                else:
                    be_not = False
                _a, cmp, _b = cls.split_calc.split(and_pattern)
                if _a and _b:
                    a = date.strftime(_a)
                    b = date.strftime(_b)
                    magic_method = getattr(a, cls.magic_methods.get(cmp, ""), None)
                    if magic_method:
                        result = magic_method(b)
                    else:
                        result = getattr(cls, cls.custom_methods[cmp])(a, b)
                    if be_not:
                        if not result:
                            continue
                        else:
                            # no match break
                            miss_pattern = miss_pattern or and_pattern
                            break
                    elif result:
                        continue
                    else:
                        # no match break
                        miss_pattern = miss_pattern or and_pattern
                        break
                else:
                    # no match break
                    miss_pattern = miss_pattern or and_pattern
                    break
            else:
                # each and_pattern is hit
                ok = True
                break
        return ok, miss_pattern
